#
# transfer.py - send and receive data
#
# Written by
#  Christian Vogelgsang <chris@vogelgsang.org>
#
# This file is part of dtv2ser.
# See README for copyright notice.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
#  02111-1307  USA.
#

import time

from dtv2ser.status  import *
from dtv2ser.cmdline import CmdLine
from dtv2ser.sercon  import SerCon
from dtv2ser.param   import Param

class Transfer:
  """Transfer command handling tools."""
  
  # how long to wait for the server after a transfer operation
  server_ready_timeout = 500
  
  # how many retries to get the transfer result after a transfer operation
  get_result_retries = 20
  
  # block size
  block_size = 0x1000
  
  def __init__(self,con):
    """Create a new transfer object."""
    # get cmdline
    self.con = con
    self.cmdline = CmdLine(con)
    # init transfer rate
    self.server_tx_rate = 0
    self.server_rx_rate = 0
    self.client_tx_rate = 0
    self.client_rx_rate = 0
    
  # ---------- transfer commands --------------------------------------------

  def get_transfer_result(self):
    """Issue a 'tr' command to get the result
    Returns result code.
    """
    result = self.cmdline.do_command("tr")
    if result != STATUS_OK:
      return result
    
    # get transfer result
    (result,transfer_result) = self.cmdline.get_byte()
    if result != STATUS_OK:
      return result
    elif transfer_result != STATUS_OK:
      return transfer_result | TRANSFER_ERROR_MASK
    else:
      return STATUS_OK
  
  def wait_for_transfer_result(self):
    """Wait until the transfer result is available or a time out occurs
    Returns result code.
    """
    # wait for CTS
    result = self.cmdline.wait_for_server(self.server_ready_timeout)
    if result != STATUS_OK:
      return result
    
    # wait a bit for the result
    for x in xrange(self.get_result_retries):
      result = self.get_transfer_result()
      # everything is ok... return
      if result == STATUS_OK:
        return STATUS_OK
      # a transfer error occured -> report it!
      if result & TRANSFER_ERROR_MASK == TRANSFER_ERROR_MASK:
        return result
      # any other error: retry
      time.sleep(0.1)
    
    # client timed out
    return CLIENT_ERROR_SERVER_TIMEOUT

  def get_transfer_crc16(self):
    """Returns crc16 checksum of last transfer with 'tc' command
    Returns (result,crc16)
    """
    result = self.cmdline.do_command("tc")
    if result != STATUS_OK:
      return (result,0)

    # get crc16
    return self.cmdline.get_word()

  def get_transfer_time(self):
    """Returns time in 10ms counts of last transfer with 'tt' command
    Returns (result,time)
    """
    result = self.cmdline.do_command("tt")
    if result != STATUS_OK:
      return (result,0)

    # get duration
    return self.cmdline.get_word()

  # ---------- transfer tools -----------------------------------------------

  def calc_crc16(self,data):
    """Calc a block checksum with crc16.
    Return crc16.
    """
    crc = 0xffff
    for a in data:
      b = ord(a)
      crc ^= b;
      for i in xrange(8):
        if crc & 1:
          crc = (crc >> 1) ^ 0xA001
        else:
          crc = (crc >> 1)
    return crc & 0xffff
  
  def compare_crc16(self,data):
    """Compare the calculated crc16 with the server's crc16 of a data block.
    Return result.
    """
    return STATUS_OK
    # do crc16 check myself
    my_crc16 = self.calc_crc16(data)
    # get crc16 from server
    (result,crc16) = self.get_transfer_crc16()
    if crc16 != my_crc16:
      return CLIENT_ERROR_CRC16_MISMATCH
    return STATUS_OK
  
  def update_server_tx_rate(self,length):
    """Get transfer time from server and calculate the tx rate.    
    Return result.
    """
    (result,server_time) = self.get_transfer_time()
    if result != STATUS_OK:
      return result
    if server_time > 0:
      self.server_tx_rate = length * 100.0 / (server_time * 1024.0)
    else:
      self.server_tx_rate = 0

    return STATUS_OK

  def update_server_rx_rate(self,length):
    """Get transfer time from server and calc transfer rate
    Return result.
    """
    (result,server_time) = self.get_transfer_time()
    if result != STATUS_OK:
      return (result,data)
    if server_time > 0:
      self.server_rx_rate = length * 100.0 / (server_time * 1024.0)
    else:
      self.server_rx_rate = 0

    return STATUS_OK
  
  def update_client_tx_rate(self,length,duration):
    """Update the server transfer rate from the 
    """
    self.client_tx_rate = length / (duration * 1024.0)

  def update_client_rx_rate(self,length,duration):
    """After a receive transfer update statistics
    """
    self.client_rx_rate = length / (duration * 1024.0)

  # ---------- block transfer commands --------------------------------------
  
  def do_send_command(self,cmd,data,callback=lambda x:True,check_crc=True):
    """Transmit a send data command and upload data.
    Return result.
    """
    # send command
    result = self.cmdline.do_command(cmd)
    if result != STATUS_OK:
      return result
      
    # upload data
    (result,duration) = self.con.send_block(data,self.block_size,callback=callback)
    if result != STATUS_OK:
      return result

    # check transfer result
    result = self.wait_for_transfer_result()
    if result != STATUS_OK:
      return result
      
    # update tx rate
    length = len(data)
    self.update_client_tx_rate(length,duration)
    result = self.update_server_tx_rate(length)
    if result != STATUS_OK:
      return result
    
    # checksum check
    if check_crc:
      return self.compare_crc16(data)  
    else:
      return STATUS_OK
  
  def do_receive_command(self,cmd,length,callback=lambda x:True,check_crc=True):
    """Transmit a receive data command and download data.
    Return (result,data)
    """
    # send command
    result = self.cmdline.do_command(cmd)
    if result != STATUS_OK:
      return (result,'')

    # download data
    (result,duration,data) = self.con.receive_block(length,self.block_size,callback=callback)
    if result != STATUS_OK:
      return result

    # check transfer result
    result = self.wait_for_transfer_result()
    if result != STATUS_OK:
      return (result,data)
    
    # update rx rate
    self.update_client_rx_rate(length,duration)
    result = self.update_server_rx_rate(length)
    if result != STATUS_OK:
      return (result,data)
    
    # checksum check
    if check_crc:
      return (self.compare_crc16(data),data)
    else:
      return (STATUS_OK,data)

  # ---------- diagnose commands --------------------------------------------

  def set_diagnose_pattern(self,pattern):
    """Set the diagnose pattern in dtv2ser.
    Return result.
    """
    param = Param(self.con)
    return param.set_byte(PARAM_BYTE_DIAGNOSE_PATTERN,pattern)
  
  def get_diagnose_pattern(self):
    """Read the diagnose pattern from dtv2ser.
    Return (result,pattern).
    """
    param = Param(self.con)
    return param.get_byte(PARAM_BYTE_DIAGNOSE_PATTERN)
  
  def do_diagnose_command_only_dtv(self,cmd,length,is_tx,callback=lambda x:True):
    """Issue a diagnose send command but do not transfer data via serial.
    Return result.
    """
    # send command
    result = self.cmdline.do_command(cmd);
    if result != STATUS_OK:
      return result
    
    # wait for result
    start_time = time.time()
    result = self.wait_for_transfer_result()
    end_time = time.time()
    if result != STATUS_OK:
      return result
    
    # update tx or rx rate
    if is_tx:
      self.update_client_tx_rate(length,end_time - start_time)
      return self.update_server_tx_rate(length)
    else:
      self.update_client_rx_rate(length,end_time - start_time)
      return self.update_server_rx_rate(length)

  def do_diagnose_send_command_only_client(self,cmd,length,callback=lambda x:True):
    """Issue a diagnose send command but do not transfer to the DTV.
    Return result.
    """
    # get pattern
    status,pattern = self.get_diagnose_pattern()
    if status != STATUS_OK:
      return status
    
    # create test data
    data = chr(pattern) * length

    # perform send command
    return self.do_send_command(cmd,data,callback=callback)
    
  def do_diagnose_receive_command_only_client(self,cmd,length,callback=lambda x:True):
    """Issue a diagnose receive command but do not transfer to the DTV.
    Return result.
    """
    # get pattern
    status,pattern = self.get_diagnose_pattern()
    if status != STATUS_OK:
      return status

    # perform receive command
    (result,data) = self.do_receive_command(cmd,length,callback=callback)
    if result != STATUS_OK:
      return result
      
    # check data
    expected_data = chr(pattern) * length
    if expected_data != data:
      return CLIENT_ERROR_CORRUPT_SERIAL_DATA
    else:
      return STATUS_OK

  # ---------- get statistics -----------------------------------------------
  
  def get_rx_rates(self):
    """Return client and server rx rates."""
    return (self.client_rx_rate,self.server_rx_rate)
  
  def get_tx_rates(self):
    """Return client and server tx rates."""
    return (self.client_tx_rate,self.server_tx_rate)
