/*
 * command.c - handle commands
 *
 * Written by
 *  Christian Vogelgsang <chris@vogelgsang.org>
 *
 * This file is part of dtv2ser.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

#include "command.h"
#include "display.h"
#include "dtvlow.h"
#include "dtvtrans.h"
#include "uart.h"
#include "sertrans.h"
#include "param.h"
#include "timer.h"

// ----- Helpers -----

// enter error condition to timeout host transfer
static void error_condition(void)
{
  u08 toggle = 1;
  led_error_on();
    
  // do some blinking
  u08 num = PARAM_BYTE(PARAM_BYTE_ERROR_CONDITION_LOOPS);
  for(u08 i=0;i<num;i++) {
    timer_delay_10ms(PARAM_WORD(PARAM_WORD_ERROR_CONDITION_DELAY));
    toggle ^= 1;
    if(toggle) {
      led_error_on();
    } else {
      led_error_off();
    }
  }

  // flush serial
  uart_clear_buffer();
  led_error_off();
}

#ifdef USE_LCD
static void lcd_print_dir(u08 to_host)
{
  static u08 *to_host_msg = (u08 *)"d>s";
  static u08 *to_dtv_msg  = (u08 *)"d<s";
  lcd_clear();
  lcd_print_string(0,0,to_host ? to_host_msg : to_dtv_msg);
}
#endif

static void generic_transfer(u08 to_host,
                             cmdline_args_t *args,
                             host_transfer_funcs_t *host_funcs,
                             dtv_transfer_block_func_t dtv_func)
{
  u08 mode = args->arg_byte[0];
  u32 addr = args->arg_dword[0];
  u32 len  = args->arg_dword[1];

#ifdef USE_LCD
  lcd_print_dir(to_host);
  lcd_print_dword6(8,0,'l',len);
  lcd_print_dword6(8,1,'a',addr);
  lcd_print_byte(0,1,'m',mode);
#endif  

  // perform transfer
  u16 block_size = PARAM_WORD(PARAM_WORD_DTV_TRANSFER_BLOCK_SIZE);
  u08 status = transfer_mem(mode,addr,len,
                            block_size,
                            host_funcs,
                            dtv_func);

#ifdef USE_LCD
  lcd_print_byte(3,0,'s',status);
#endif

  if(status!=TRANSFER_OK) {
    error_condition();
  }
}

#ifdef USE_BLOCKCMD

static void generic_block_transfer(u08 to_host,
                                   cmdline_args_t *args,
                                   host_transfer_funcs_t *host_funcs,
                                   dtv_transfer_block_func_t dtv_func)
{
  u08 mode   = args->arg_byte[0];
  u08 bank   = args->arg_byte[1];
  u16 offset = args->arg_word[0];
  u16 length = args->arg_word[1];

#ifdef USE_LCD
  lcd_print_dir(to_host);
  lcd_print_word(10,0,'l',length);
  lcd_print_word(10,1,'o',offset);
  lcd_print_byte(0,1,'m',mode);
  lcd_print_byte(5,1,'b',bank);
#endif  
  
  // perform transfer
  u08 status = transfer_mem_block(mode,bank,offset,length,
                                  host_funcs,
                                  dtv_func);
  
#ifdef USE_LCD
  lcd_print_byte(5,0,'s',status);
#endif
  
  if(status!=TRANSFER_OK) {
    error_condition();
  }
}

#endif

// ---------- Commands ------------------------------------------------------

// ----- Read/Write Memory -----

static u08 validate_read_memory(cmdline_args_t *args)
{
  u32 base = args->arg_dword[0];
  u32 len  = args->arg_dword[1];
  if(base+len > 0x200000)
    return CMDLINE_ERROR_VALIDATION;
  if(args->arg_byte[0]>1)
    return CMDLINE_ERROR_VALIDATION;  
  return CMDLINE_STATUS_OK;
}

static u08 validate_write_memory(cmdline_args_t *args)
{
  u08 status = validate_read_memory(args);
  if(status==CMDLINE_STATUS_OK) {
    if(args->arg_byte[0]==0) // must be mode ram
      return CMDLINE_STATUS_OK;
    else
      return CMDLINE_ERROR_VALIDATION;
  } else
    return status;
}

static void exec_read_memory(cmdline_args_t *args)
{
  generic_transfer(1,args,&serial_host_write_funcs,dtvtrans_recv_mem_block);
}

static void exec_write_memory(cmdline_args_t *args)
{
  generic_transfer(0,args,&serial_host_read_funcs,dtvtrans_send_mem_block);
}

#ifdef USE_DIAGNOSE

static void exec_read_memory_only_dtv(cmdline_args_t *args)
{
  diagnose_host_mode = DIAGNOSE_HOST_MODE_WRITE;
  generic_transfer(1,args,&diagnose_host_transfer_funcs,dtvtrans_recv_mem_block);
}

static void exec_write_memory_only_dtv(cmdline_args_t *args)
{
  diagnose_host_mode = DIAGNOSE_HOST_MODE_READ;
  generic_transfer(0,args,&diagnose_host_transfer_funcs,dtvtrans_send_mem_block);
}

static void exec_verify_memory_only_dtv(cmdline_args_t *args)
{
  diagnose_host_mode = DIAGNOSE_HOST_MODE_VERIFY;  
  generic_transfer(1,args,&diagnose_host_transfer_funcs,dtvtrans_recv_mem_block);
}

#endif // USE_DIAGNOSE

// ----- Serial Diagnose -----

#ifdef USE_DIAGNOSE

static void exec_read_memory_only_serial(cmdline_args_t *args)
{
  generic_transfer(1,args,&serial_host_write_funcs,diagnose_dtv_recv_block);
}

static void exec_write_memory_only_serial(cmdline_args_t *args)
{
  generic_transfer(0,args,&serial_host_read_funcs,diagnose_dtv_send_block);  
}

#endif // USE_DIAGNOSE

// ----- Go Memory -----

static void exec_go_memory(cmdline_args_t *args)
{
  u16 addr = args->arg_word[0];
  
#ifdef USE_LCD
  lcd_clear();
  lcd_print_string(3,1,(u08*)"goto");
  lcd_print_word(8,1,'a',addr);
#endif

  // perform go
  led_transmit_on();
  u08 status = dtvtrans_exec_mem(addr);
  led_transmit_off();

  uart_send_hex_byte_crlf(status);
}

// ----- Reset DTV -----

static void exec_reset_dtv(cmdline_args_t *args)
{
#ifdef USE_LCD
  lcd_clear();
  lcd_print_string(0,1,(u08*)"reset");
#endif

  // perform reset
  led_transmit_on();
  dtvlow_reset_dtv();
  led_transmit_off();
  
  // send result
  uart_send_hex_byte_crlf(0);
}

// ----- Version -----

static void exec_version(cmdline_args_t *args)
{
#ifdef USE_LCD
  lcd_clear();
  lcd_print_version();
#endif

  // return version number of dtv2ser
  uart_send_hex_word_crlf(VERSION_MAJ << 8 | VERSION_MIN);
}

// ----- Transfer State -----

static void exec_transfer_result(cmdline_args_t *args)
{
  uart_send_hex_byte_crlf(transfer_state.result);
}

static void exec_transfer_length(cmdline_args_t *args)
{
  uart_send_hex_word_crlf(transfer_state.length);
}

static void exec_transfer_time(cmdline_args_t *args)
{
  uart_send_hex_word_crlf(transfer_state.ms_time);
}

// ----- Boot Memory -----

#ifdef USE_BOOTSTRAP

static void exec_boot_memory(cmdline_args_t *args)
{
  u16 addr = args->arg_word[0];
  u16 len  = args->arg_word[1];
  
  // perform transfer
  u08 status = transfer_mem(0,addr,len,
                            0x4000,
                            &serial_host_read_funcs,
                            dtvtrans_send_mem_block_boot);
  
  if(status!=TRANSFER_OK)
    error_condition();
}

#endif

// ----- Read/Write Block Memory -----

#ifdef USE_BLOCKCMD

static u08 validate_block_read_memory(cmdline_args_t *args)
{
  u08 mode = args->arg_byte[0];
  if(mode>1)
    return CMDLINE_ERROR_VALIDATION;
  
  u16 offset = args->arg_word[0];
  if(offset>=0x4000)
    return CMDLINE_ERROR_VALIDATION;

  u16 length = args->arg_word[1];
  if((offset+length)>0x4000)
    return CMDLINE_ERROR_VALIDATION;
    
  return CMDLINE_STATUS_OK;
}

static u08 validate_block_write_memory(cmdline_args_t *args)
{
  u08 status = validate_block_read_memory(args);
  if(status==CMDLINE_STATUS_OK) {
    if(args->arg_byte[0]==0) // must be mode ram
      return CMDLINE_STATUS_OK;
    else
      return CMDLINE_ERROR_VALIDATION;
  } else
    return status;
}

static void exec_block_read_memory(cmdline_args_t *args)
{
  generic_block_transfer(1,args,&serial_host_write_funcs,dtvtrans_recv_mem_block);
}

static void exec_block_write_memory(cmdline_args_t *args)
{
  generic_block_transfer(0,args,&serial_host_read_funcs,dtvtrans_send_mem_block);
}

#endif

#ifdef USE_DIAGNOSE

#ifdef USE_BLOCKCMD

static void exec_block_read_memory_only_dtv(cmdline_args_t *args)
{
  generic_block_transfer(1,args,&diagnose_host_transfer_funcs,dtvtrans_recv_mem_block);
}

static void exec_block_write_memory_only_dtv(cmdline_args_t *args)
{
  generic_block_transfer(0,args,&diagnose_host_transfer_funcs,dtvtrans_send_mem_block);
}

#endif

static void trigger_error_condition(cmdline_args_t *args)
{
  error_condition();
}

#endif // USE_DIAGNOSE

// ----- Parameter Commands -----

static u08 validate_byte_param(cmdline_args_t *args)
{
  if(args->arg_byte[0] < PARAM_BYTE_MAX)
    return CMDLINE_STATUS_OK;
  else
    return CMDLINE_ERROR_VALIDATION;
}

static u08 validate_word_param(cmdline_args_t *args)
{
  if(args->arg_byte[0] < PARAM_WORD_MAX)
    return CMDLINE_STATUS_OK;
  else
    return CMDLINE_ERROR_VALIDATION;
}

static void exec_set_byte_param(cmdline_args_t *args)
{
  PARAM_BYTE(args->arg_byte[0]) = args->arg_byte[1];
  uart_send_hex_byte_crlf(0);
}

static void exec_get_byte_param(cmdline_args_t *args)
{
  u08 value = PARAM_BYTE(args->arg_byte[0]);
  uart_send_hex_byte_crlf(value);
}

static void exec_set_word_param(cmdline_args_t *args)
{
  PARAM_WORD(args->arg_byte[0]) = args->arg_word[0];
  uart_send_hex_byte_crlf(0);
}

static void exec_get_word_param(cmdline_args_t *args)
{
  u16 value = PARAM_WORD(args->arg_byte[0]);
  uart_send_hex_word_crlf(value);
}

static void exec_save_param(cmdline_args_t *args)
{
  u08 result = param_save();
  uart_send_hex_byte_crlf(result);
}

static void exec_load_param(cmdline_args_t *args)
{
  u08 result = param_load();
  uart_send_hex_byte_crlf(result);
}

static void exec_reset_param(cmdline_args_t *args)
{
  param_reset();
  uart_send_hex_byte_crlf(0);
}

static void exec_dump_byte_params(cmdline_args_t *args)
{
  uart_send_hex_byte_crlf(PARAM_BYTE_MAX);
  for(u08 i=0;i<PARAM_BYTE_MAX;i++) {
    uart_send_hex_byte_crlf(PARAM_BYTE(i));
  }
}

static void exec_dump_word_params(cmdline_args_t *args)
{
  uart_send_hex_byte_crlf(PARAM_WORD_MAX);
  for(u08 i=0;i<PARAM_WORD_MAX;i++) {
    uart_send_hex_word_crlf(PARAM_WORD(i));
  }
}

static void exec_dump_params(cmdline_args_t *args)
{
  exec_dump_byte_params(args);
  exec_dump_word_params(args);
}

// ---------- The Command Table ---------------------------------------------
command_t command_table[] = {
  // main commands
  COMMAND("r","btt",validate_read_memory,exec_read_memory),
  COMMAND("w","btt",validate_write_memory,exec_write_memory),
  COMMAND("x",0,0,exec_reset_dtv),
  COMMAND("g","w",0,exec_go_memory),
  COMMAND("v",0,0,exec_version),
#ifdef USE_BOOTSTRAP
  COMMAND("b","ww",0,exec_boot_memory),
#endif
  
  // transfer status commands
  COMMAND("tr",0,0,exec_transfer_result),
  COMMAND("tl",0,0,exec_transfer_length),
  COMMAND("tt",0,0,exec_transfer_time),

  // preset commands
  COMMAND("pbs","bb",validate_byte_param,exec_set_byte_param),
  COMMAND("pbg","b",validate_byte_param,exec_get_byte_param),
  COMMAND("pws","bw",validate_word_param,exec_set_word_param),
  COMMAND("pwg","b",validate_word_param,exec_get_word_param),
  COMMAND("ps",0,0,exec_save_param),
  COMMAND("pl",0,0,exec_load_param),
  COMMAND("pr",0,0,exec_reset_param),
  COMMAND("pbd",0,0,exec_dump_byte_params),
  COMMAND("pwd",0,0,exec_dump_word_params),
  COMMAND("pd",0,0,exec_dump_params),
  
  // block commands
#ifdef USE_BLOCKCMD
  COMMAND("dbr","bbww",validate_block_read_memory,exec_block_read_memory),
  COMMAND("dbw","bbww",validate_block_write_memory,exec_block_write_memory),
#endif

  // diagnose commands
#ifdef USE_DIAGNOSE
  COMMAND("ddr","btt",validate_read_memory,exec_read_memory_only_dtv),
  COMMAND("ddw","btt",validate_write_memory,exec_write_memory_only_dtv),
  COMMAND("ddv","btt",validate_read_memory,exec_verify_memory_only_dtv),
#ifdef USE_BLOCKCMD
  COMMAND("ddbr","bbww",validate_block_read_memory,exec_block_read_memory_only_dtv),
  COMMAND("ddbw","bbww",validate_block_write_memory,exec_block_write_memory_only_dtv),
#endif
  COMMAND("dsr","btt",validate_read_memory,exec_read_memory_only_serial),
  COMMAND("dsw","btt",validate_write_memory,exec_write_memory_only_serial),
  COMMAND("de",0,0,trigger_error_condition),
#endif
  
  END_OF_COMMAND
};

